% Pipeline for supervised learning (classification) using libSVM

% Requires libSVM to be downloaded/compiled/added to path

% Input 
%   dcmlist - data (format depends on the dataType)
% 
% settings
%   kernelType = 0 for linear , 1 for gaussian kernel 
%   dataType = 1 for DCMs, 2 for HGF, 0 for (X,y)  
%   toolType = 1 for SVM, 2 for lasso GLM
%   extraFeatures - matrix of features apart from provided or derived from
%   crossval = -1 for leave one out , and  k for k-fold crossvalidation
%   dcmlist
%
% Author - Sudhir Shankar Raman, TNU , 2015


function [output] = supervised_learning_pipeline(dcmlist,settings)

kernelType = settings.kernelType;
crossV = settings.crossV;
% load data in the appropriate format

X = dcmlist.X;
y = dcmlist.y;

if(crossV == -1)
    crossV = length(y);
end


% do analysis with cross validation
% classify using linear kernel

paramList = create_param_list(kernelType);
scores = cross_validate(X,y,paramList,kernelType,1,crossV);
[bestParam,score] = choose_param(scores,paramList);
output = struct();
output.bestParam = bestParam;
output.score = score;    
output.scores = scores;
output.paramList = paramList;



% Normalize features based on settings.iNorm and settings.iStd

function [paramStr] = constructParams(param,kernelType,toolType)

%paramstr = '-c 1 -g 0.02 -b 1';
switch toolType
    case 1
        switch kernelType
            case 2
                paramStr = ['-t 2 -s 0 -c ',num2str(param(1)) ,' -g ',num2str(param(2)) ,' -b 1 -q'];
            otherwise
                paramStr = ['-t 0 -s 0 -c ',num2str(param(1)) ,' -b 1 -q'];
        end
    otherwise
        paramStr = ['-s 0 -c ',num2str(param(1)) ,' -q'];
end

function [paramList] = create_param_list(kernelType)


switch kernelType
    case 2
        % Gaussian kernel
        paramList = [2.^[-6:7]' 2.^[-9:4]'];
        
    otherwise
        % Linear kernel
        paramList = 2.^[-6:6]';
end

return;

function [bestParam,score] = choose_param(scores,paramList)

[score,max_idx] = max(scores(:));

[i,j]=ind2sub(size(scores),max_idx);
if(size(scores,2) ~= 1)
    bestParam = [paramList(i,1) paramList(j,2)]';
else
    bestParam = [paramList(i,1)]';
end
return;

function [scores] = cross_validate(X,y,paramList,kernelType,toolType,crossV)

scores = zeros(size(paramList));
switch kernelType
    case 2
        for i=1:length(paramList(:,1))   
            for j=1:length(paramList(:,2))
                [paramStr] = constructParams([paramList(i,1) paramList(j,2)],kernelType,toolType);
                scores(i,j) = classify(X,y,paramStr,toolType,crossV);        
            end
        end
    otherwise
        for i=1:size(paramList,1)   
           [paramStr] = constructParams([paramList(i,1)],kernelType,toolType);
           scores(i) = classify(X,y,paramStr,toolType,crossV);
        end
end

function [score] = classify(X,y,paramstr,toolType,crossV)

label = y;
data = X;

N = length(data);

NumofClass = length(unique(label));

% Create fold indices

startind =  floor(1:(N/crossV):N);
endind=startind;
endind(1)=[];
endind=endind-1;
endind(end+1)=N;

predLabel = [];
testLabel1 = [];
%for iSubject = 1:length(data)
for iFold = 1:length(startind)
        
    % calculated indices for training and test data
    trainingLabel = label;
    trainingLabel(startind(iFold):endind(iFold),:) = [];
    trainingData = data;
    trainingData(startind(iFold):endind(iFold),:) = [];
    
    % Training
    model = cell(NumofClass,1); 
    for k = 1:NumofClass        
        switch toolType
            case 1
                model{k} = svmtrain(double(trainingLabel==k), trainingData,paramstr);        
            otherwise
                model{k} = train(double(trainingLabel==k), sparse(trainingData),paramstr);        
        end
    end

    %% calculate the probability of different labels
    testLabel = label(startind(iFold):endind(iFold));
    testData = data(startind(iFold):endind(iFold),:);

    pr = [];
    for k = 1:NumofClass
         switch toolType
            case 1
                [a,b,p] = svmpredict(double(testLabel==k), testData, model{k}, '-b 1 -q');
                pr(:,k) = p(:,model{k}.Label==1);    %# probability of class==k
            
             otherwise
                 [a,b,p] = predict(double(testLabel==k), sparse(testData), model{k}, '-b 1 -q');
                pr(:,k) = p(:,model{k}.Label==1);    %# probability of class==k
         end
    end

    %% label prediction will be the one with highest probability:

    [a,predctedLabel] = max(pr,[],  2);

    %predLabel(iFold,1) = predctedLabel(1);
    predLabel = [predLabel ;predctedLabel];
    testLabel1 = [testLabel1;testLabel];

end

% Compute balanced accuracy
score = compute_accuracy(predLabel,testLabel1,2);

function [X,y] = filter_class(X,y,k)

[a,b]=find(y == k);
X(a,:) = [];         
y(a) = [];
y = floor(y/max(y));
y = y+1;


function [score] = compute_accuracy(predLabel,truelabel,type)

switch type
    case 2
        % balanced accuracy
        nclass = length(unique(truelabel));
        score = 0;
        for i = 1:nclass
            trueclass = find(truelabel == i);
            posres = predLabel(trueclass) - truelabel(trueclass);
            TP = length(find(posres == 0));
            totalP = length(posres);
            score = score + (TP/(totalP));
        end
        score = 100*(1/nclass)*score;
    otherwise
        res=truelabel - (predLabel);
        ncorrect=sum(res == 0);
        accuracy = 100*ncorrect/length(predLabel);
        score = accuracy;
end

